% paraperspective factorization
% description: Paraperspective factorization
%
% Inputs:
% rpy_ix: x coordinate of feature points in images [no. of feature, no. of images] = size(rpy_ix)
% rpy_iy: y coordinate of feature points in images
% flen: focal length in no. of pixels
% init_center: initial center point of the 3D model
% noit: no use (not implemented)
% 
% Outputs:
% P: found model
% RT_sol: Translation [1:3] and Rotation angle [4:6]
% error_trucco_all_time: (not implemented)
% mse_error_model_pixels2_loop: (not implemented)
function [P,RT_sol,error_trucco_all_time,mse_error_model_pixels2_loop]...
    = paraperspectivefac(rpy_ix,rpy_iy,flen,init_center,centered_real_model,noit)

[nf,Nimg] = size(rpy_ix);

% find X and Y for W
X = rpy_ix' / flen;
Y = rpy_iy' / flen;

% paraperspective factorization
[I,J,K,T,S] = paraperspectivefac(X,Y);

% find correct model (rotated and translated)
firstCamDirection = [I(1,:);J(1,:);K(1,:)];
Ph = firstCamDirection*S;
%Ph = Ph + (T(1,:))'*ones(1,nf);

% correspond to centroid
cPh = (sum(Ph'))'/nf;
rPep = zeros(3,nf);
rP = zeros(3,nf);
rPh = Ph - cPh * ones(1,nf);

% scaling factor
avePh = sum(sum(abs(rPh)));
aveP = sum(sum(abs(centered_real_model)));
rPh = rPh * aveP / avePh;
%rPh(3,:) = -rPh(3,:);

P = rPh;

% Translation
RT_sol = zeros(6,Nimg);
Th = T * aveP / avePh; % scaling factor
RT_sol(1:3,:) = Th';

% Rotation
for i=1:Nimg
    oi = I(i,:)';
    oj = J(i,:)';
    ok = K(i,:)';
    rotation_matrix = inv(firstCamDirection) * [oi oj ok]';
    RT_sol(4:6,i) = -rpyAng(rotation_matrix);
end

error_trucco_all_time = zeros(noit,1);
mse_error_model_pixels2_loop = zeros(noit,1);





% X,Y
% R,S
function [I,J,K,T,S] = paraperspectivefac(X,Y)

% N: No of images
% n: No of feature pts
[N,n] = size(X);

% generate ~W
% s_W: xf and yf
W = zeros(2*N,n);
W(1:N,:) = X;
W(N+1:2*N,:) = Y;
s_W = (sum(W'))' / n;
%for i = 1:n
%    W(:,i) = W(:,i) - s_W;
%end
W = W - s_W * ones(1,n);

% svd
[U,D,V] = svds(W,3);
D2 = sqrtm(D);
RH = U*D2;
SH = D2*V';
%RH = U;
%SH = D*V';

% normalization
% creating Z & c, Zl=c
Z = zeros(2*N+1,6);
for i=1:N
    mh = (RH(i,:))';
    nh = (RH(i+N,:))';
    xf = s_W(i);
    yf = s_W(N+i);
    a = 1 + xf*xf;
    b = 1 + yf*yf;
    c = 2*a*b / (xf * yf);
    gm = g(mh,mh);
    gn = g(nh,nh);
    gmn = g(mh,nh);
    Z(i,:) = (b*gm+a*gn-c*gmn)';
    Z(N+i,:) = (b*gm-a*gn)';
end
Z(2*N+1,:) = (g(RH(1,:)',RH(1,:)'))';

c = zeros(2*N+1,1);
c(2*N+1) = 1;

% find sl (vector l)
sl = pinv(Z'*Z)*Z'*c;

% form L
L = [sl(1) sl(2) sl(3);sl(2) sl(4) sl(5);sl(3) sl(5) sl(6)];

% eigendecomposition
[U,D,V] = svd(L);
Q = U * sqrtm(D);    % ?????????????

% M & S
M = RH * Q;
S = inv(Q) * SH;

% Create Gf and Hf
% Create R
I = zeros(N,3);
J = zeros(N,3);
K = zeros(N,3);
T = zeros(N,3);
Gf = zeros(3,3);
Hf = ones(3,1);
for i=1:N
    xf = s_W(i);
    yf = s_W(N+i);
    msf = M(i,:) * sqrt(1+xf*xf) / norm(M(i,:));
    nsf = M(i+N,:) * sqrt(1+yf*yf) / norm(M(i+N,:));
    Gf(1,:) = cross(msf,nsf);
    Gf(2,:) = msf;
    Gf(3,:) = nsf;
    Hf(2) = -xf;
    Hf(3) = -yf;
    k_f = (inv(Gf) * Hf)';
    i_f = cross(nsf,k_f);
    j_f = cross(k_f,msf);
    I(i,:) = i_f;
    J(i,:) = j_f;
    K(i,:) = k_f;

%     dot(i_f,i_f)
%     dot(j_f,j_f)
%     dot(i_f,j_f)
%     dot(k_f,k_f)
    
    % find tf
    zf = (sqrt((1+xf*xf)/dot(M(i,:),M(i,:)))+sqrt((1+yf*yf)/dot(M(N+i,:),M(N+i,:))))/2;
%    zf = sqrt((1+xf*xf)/dot(M(i,:),M(i,:)));
    tf = -zf*([xf yf 1]*inv([i_f;j_f;k_f]'));
    T(i,:) = tf;
end



% function g
% a = [a1*b1 a1*b2+a2*b1 a1*b3+a3*b1 a2*b2 a2*b3+a3*b2 a3*b3]'
function a = g(af,bf)
a = [af(1)*bf(1) af(1)*bf(2)+af(2)*bf(1) af(1)*bf(3)+af(3)*bf(1) af(2)*bf(2) af(2)*bf(3)+af(3)*bf(2) af(3)*bf(3)]';

