% sequentialfac
% main method for sequential factorization (all frames)
% For CT
%
% Inputs:
% rpy_ix: x coordinate of feature points in images [no. of feature, no. of images] = size(rpy_ix)
% rpy_iy: y coordinate of feature points in images
% flen: focal length in no. of pixels
% init_center: initial center point of the 3D model
% noit: no use(not implemented)
% 
% Outputs:
% P: found model
% RT_sol: Translation [1:3] and Rotation angle [4:6]
% error_trucco_all_time: (not implemented)
% mse_error_model_pixels2_loop: (not implemented)
function [P,RT_sol,error_trucco_all_time,mse_error_model_pixels2_loop]...
    = sequentialfac_ct(rpy_ix,rpy_iy,flen,init_center,noit)

[nf,Nimg] = size(rpy_ix);

% find X and Y for W
X = rpy_ix' * init_center(3) / flen;
Y = rpy_iy' * init_center(3) / flen;

% sequential factorization
[R,S] = sequentialfac(X,Y);

% find the Direction of first Cam (respect to object centroid)
oi = R(1,:)';
oj = R(Nimg+1,:)';
ok=cross(oi,oj);
firstCamDirection = [oi oj ok]';

P = firstCamDirection*S;

% Rotation and translation (redundant for original factorization)
RT_sol = zeros(6,Nimg);
RT_sol(1:3,1) = init_center;
for i=1:Nimg
    oi = R(i,:)';
    oj = R(Nimg+i,:)';
    ok = cross(oi,oj);
    rotation_matrix = inv(firstCamDirection) * [oi oj ok]';
    RT_sol(4:6,i) = rpyAng(rotation_matrix);
end

error_trucco_all_time = zeros(noit,1);
mse_error_model_pixels2_loop = zeros(noit,1);





% sequential factorization (core method)
function [R,S] = sequentialfac(X,Y)
[Nimg, nf] = size(X);
R = zeros(Nimg*2,3);

Zf1 = zeros(nf);
Qf1 = zeros(nf,3);
Qf1(1:3,:) = Qf1(1:3,:) + eye(3);
QHf1 = zeros(nf,3);
QHf1(1:3,:) = QHf1(1:3,:) + eye(3);
Df1 = zeros(6);
Ef1 = zeros(6,1);

for i = 1:Nimg
%    display(i);
    xf = (X(i,:))';
    yf = (Y(i,:))';
    [Sf,iff,jff,Zf1,Qf1,QHf1,Df1,Ef1] = sequentialfac_each(xf,yf,Zf1,Qf1,QHf1,Df1,Ef1);
    R(i,:) = iff';
    R(Nimg+i,:) = jff';
end

S = Sf;



% sequentialfac_each
% Sequential factorization for each frame
% xf: [x1 x2 ... x(nf)]' for frame f
% yf: [y1 y2 ... y(nf)]' for frame f
% Zf1: Z(f-1), Z in previous frame (zero matrix for 1st) [nf x nf]
% Qf1: Q(f-1), Qf in previous frame (must be orthonormal column) [nf x 3] 
% QHf1: /Q(f-1), QHf in previous frame (must be orthonormal column) [nf x 3]
% Df1: D(f-1), [6 x 6]
% Ef1: E(f-1), [6 x 1]
%
% Sf: Shape after this frame
% iff: [ix iy iz]'
% jff: [jx jy jz]'
% Zf: Z in this frame
% Qf
% QHf
% Df
% Ef
function [Sf,iff,jff,Zf,Qf,QHf,Df,Ef] = sequentialfac_each(xf,yf,Zf1,Qf1,QHf1,Df1,Ef1)

% algorithm 1
Zf = Zf1 + xf * xf' + yf * yf';
Y = Zf * Qf1;
[Qf, R] = qr(Y,0);  % QR (with n column only)

% algorithm 2
Hf = Qf * Qf';
Y = Hf * QHf1;
[QHf,R] = qr(Y,0);

% Zf = Zf1 + xf * xf' + yf * yf';
% [Qf, tempD] = eigs(Zf,3);
% 
% % algorithm 2
% Hf = Qf * Qf';
% Y = Hf * QHf1;
% [QHf,R] = qr(Y,0);


% algorithm 3
iHf = (xf' * QHf)';
jHf = (yf' * QHf)';
gi = g(iHf,iHf);
gj = g(jHf,jHf);
gij = g(iHf,jHf);
Df = Df1 + gi*gi' + gj*gj' + gij*gij';
Ef = Ef1 + gi + gj;

% eigendecomposition
slf = pinv(Df) * Ef;
Lf = [slf(1) slf(2) slf(3);slf(2) slf(4) slf(5);slf(3) slf(5) slf(6)];
[U,D,V] = svd(Lf);
Af = U * sqrtm(D);

% finally
iff = (iHf' * Af)';
jff = (jHf' * Af)';
Sf = pinv(Af) * QHf';



function a = g(af,bf)
a = [af(1)*bf(1) af(1)*bf(2)+af(2)*bf(1) af(1)*bf(3)+af(3)*bf(1) af(2)*bf(2) af(2)*bf(3)+af(3)*bf(2) af(3)*bf(3)]';